#include <iostream>
#include <vector>
#include <fstream>

#include "ObjLoader.h"
#include "VectorMath.h"

bool CObjFile::GetFilePath(char *outstr, const char *instr) {
	int x;
	int last_slash = -1;
	int len;
	len = strlen(instr);

	for (x=0;x<len;x++) {
		if (instr[x]=='/' || instr[x]=='\\') last_slash = x;
	}
	if (last_slash<0) {
		strcpy(outstr, "./");
		return true;
	} else {
		memcpy(outstr,instr,last_slash+1);
		outstr[last_slash+1]='\0';
		return true;
	}
}

int strpos(char *needle, char *haystack)
{
	int nlen, hlen, cpos = 0;
	int x;
	bool ismatch = false;
	nlen = strlen(needle);
	hlen = strlen(haystack);

	while (cpos+nlen <= hlen) {
		ismatch = true;
		for (x=0;(x<nlen)&&ismatch;x++) {
			if (needle[x] != haystack[x+cpos]) ismatch = false;;
		}
		if (ismatch) return cpos;
		cpos++;
	}

	return -1;

}

int GetToken(char *input, char *output, int startpos)
{
	int ilen = strlen(input);
	int x = 0;
	while (input[startpos] == ' ' || input[startpos] == '\n' || input[startpos] == '\t' || input[startpos] == '\0') {
		startpos ++;
		if (startpos >= ilen) return -1;
	}
	while (input[startpos] != ' ' && input[startpos] != '\n' && input[startpos] != '\t' && input[startpos] != '\0') {
		output[x] = input[startpos];
		x++;
		startpos++;
	}
	output[x] = '\0';
	return startpos;
}

void CObjGroup::AddTri(unsigned int p1, unsigned int p2, unsigned int p3, 
					   unsigned int tc1, unsigned int tc2, unsigned int tc3,
					   unsigned int n1, unsigned int n2, unsigned int n3)
{
	SObjTriPtr NewTri;
	NewTri = new SObjTri;
	NewTri->mtl = curr_mtl;
	NewTri->p1 = p1;
	NewTri->p2 = p2;
	NewTri->p3 = p3;
	NewTri->tc1 = tc1;
	NewTri->tc2 = tc2;
	NewTri->tc3 = tc3;
	NewTri->n1 = n1;
	NewTri->n2 = n2;
	NewTri->n3 = n3;
	tri.push_back(NewTri);
}

int CObjFile::SetGroup(char *linebuff, int p)
{
	char tstr[256];
	p = GetToken(linebuff, tstr, p);
	unsigned int x;
	CObjGroupPtr NewGroup;

	if (strcmp(tstr, "default") != 0) {
		x = 0;
		while (x<list_g.size()) {
			if (strcmp(list_g[x]->sName, tstr)==0) {
				return x;
			}
			x++;
		}
		NewGroup = new CObjGroup;
		strcpy(NewGroup->sName, tstr);
		list_g.push_back(NewGroup);
		return x;
	} else {
		return 0;
	}
}

void CObjFile::ReadVertex(char *linebuff, int p)
{
	char v1[256], v2[256], v3[256];
	SVect3dPtr vect;
	p = GetToken(linebuff, v1, p);
	p = GetToken(linebuff, v2, p);
	p = GetToken(linebuff, v3, p);

	vect = new SVect3d;
	*vect = v3d((float)atof(v1), (float)atof(v2), (float)atof(v3));
	list_v.push_back(vect);
}

void CObjFile::ReadTexCoord(char *linebuff, int p)
{
	char v1[256], v2[256];
	SVect3dPtr vect;
	p = GetToken(linebuff, v1, p);
	p = GetToken(linebuff, v2, p);

	vect = new SVect3d;
	*vect = v3d((float)atof(v1), (float)atof(v2), 0.0f);
	list_vt.push_back(vect);
}

void CObjFile::ReadNormal(char *linebuff, int p)
{
	char v1[256], v2[256], v3[256];
	SVect3dPtr vect;
	p = GetToken(linebuff, v1, p);
	p = GetToken(linebuff, v2, p);
	p = GetToken(linebuff, v3, p);

	vect = new SVect3d;
	*vect = v3d((float)atof(v1), (float)atof(v2), (float)atof(v3));
	list_vn.push_back(vect);
}

void CObjFile::ReadPolygon(char *linebuff, int p, unsigned int curr_mtl, unsigned int curr_group)
{
	char tstr[256];
	char temp_v[256], temp_vt[256], temp_vn[256];
	int x, y, len, pos2;

	unsigned int p1, p2, p3, tc1, tc2, tc3, n1, n2, n3;

	p = GetToken(linebuff, tstr, p);
	len = strlen(tstr);
	// fix missing texcoord problem
	for (x=1;x<len;x++) {
		if (tstr[x]=='/' && tstr[x-1]=='/') {
			// this is a missing coord
			tstr[len+1]='\0';
			for (y=len;y>x;y--) {
				tstr[y] = tstr[y-1];
			}
			++len;
			tstr[x] = '0';
		}
	}
	for (x=0;x<len;x++) if (tstr[x]=='/') tstr[x]=' ';
	pos2 = GetToken(tstr, temp_v, 0);
	pos2 = GetToken(tstr, temp_vt, pos2);
	pos2 = GetToken(tstr, temp_vn, pos2);
	p1 = atoi(temp_v)-1;
	tc1 = atoi(temp_vt)-1;
	n1 = atoi(temp_vn)-1;

	p = GetToken(linebuff, tstr, p);
	len = strlen(tstr);
	// fix missing texcoord problem
	for (x=1;x<len;x++) {
		if (tstr[x]=='/' && tstr[x-1]=='/') {
			tstr[len+1]='\0';
			// this is a missing coord
			for (y=len;y>x;y--) {
				tstr[y] = tstr[y-1];
			}
			++len;
			tstr[x] = '0';
		}
	}
	for (x=0;x<len;x++) if (tstr[x]=='/') tstr[x]=' ';
	pos2 = GetToken(tstr, temp_v, 0);
	pos2 = GetToken(tstr, temp_vt, pos2);
	pos2 = GetToken(tstr, temp_vn, pos2);
	p2 = atoi(temp_v)-1;
	tc2 = atoi(temp_vt)-1;
	n2 = atoi(temp_vn)-1;

	p = GetToken(linebuff, tstr, p);
	while (p>0) {
		len = strlen(tstr);
		// fix missing texcoord problem
		for (x=1;x<len;x++) {
			if (tstr[x]=='/' && tstr[x-1]=='/') {
				tstr[len+1]='\0';
				// this is a missing coord
				for (y=len;y>x;y--) {
					tstr[y] = tstr[y-1];
				}
				++len;
				tstr[x] = '0';
			}
		}
		// read coords
		for (x=0;x<len;x++) if (tstr[x]=='/') tstr[x]=' ';
		pos2 = GetToken(tstr, temp_v, 0);
		pos2 = GetToken(tstr, temp_vt, pos2);
		pos2 = GetToken(tstr, temp_vn, pos2);
		p3 = atoi(temp_v)-1;
		tc3 = atoi(temp_vt)-1;
		n3 = atoi(temp_vn)-1;

		list_g[curr_group]->curr_mtl = curr_mtl;
		list_g[curr_group]->AddTri(p1, p2, p3, tc1, tc2, tc3, n1, n2, n3);

		p2 = p3;
		tc2 = tc3;
		n2 = n3;
		p = GetToken(linebuff, tstr, p);
		
	}
}

void CObjFile::ReadFile(const char *filename)
{
	std::ifstream myFile;
	char lineholder[1024];

	char directive[256];
	char tstr[256];
	int p=0;
	unsigned int x = 0;
	int curr_group = 0;
	int curr_mtl = 0;
	char FilePath[1024];
	char MtlFileName[1024];

	GetFilePath(FilePath, filename);
	myFile.open(filename);
	if (myFile.is_open()) {
		std::cout << "Opened file " << filename << "\n";

		myFile.getline(lineholder, 1024);
		while (!myFile.eof()) {

			p = GetToken(lineholder, directive);
			if (p==-1) directive[0] = '\0';

			if (strcmp(directive, "mtllib")==0) {
				// material library
				p = GetToken(lineholder, tstr, p);
				sprintf(MtlFileName, "%s%s", FilePath, tstr);
				ReadMtlFile(MtlFileName);

			} else if (strcmp(directive, "g")==0 || strcmp(directive,"o")==0) {
				// group ... if default we ignore
				// also, object ... handle as a group
				curr_group = SetGroup(lineholder, p);

			} else if (strcmp(directive, "v")==0) {
				// vertex ... add to list
				ReadVertex(lineholder, p);

			} else if (strcmp(directive, "vt")==0) {
				// tex coord ... add to list
				ReadTexCoord(lineholder, p);

			} else if (strcmp(directive, "vn")==0) {
				// normal ... add to list
				ReadNormal(lineholder, p);

			} else if (strcmp(directive, "f")==0) {
				// geometry ... read and add
				ReadPolygon(lineholder, p, curr_mtl, curr_group);

			} else if (strcmp(directive, "usemtl")==0) {
				// material selection
				p = GetToken(lineholder, tstr, p);
				for (x=0;x<list_mtl.size();x++) {
					if (strcmp(list_mtl[x]->sName, tstr)==0) {
						curr_mtl = x;
					}
				}

			} else {
			}
			myFile.getline(lineholder, 1024);
		}

		myFile.close();
	} else {
		std::cout << "Unable to open file.\n";
	}
}

SVect3d GrabVect(char *linebuff, int p)
{
	char v1[256], v2[256], v3[256];
	SVect3d vect;
	p = GetToken(linebuff, v1, p);
	p = GetToken(linebuff, v2, p);
	p = GetToken(linebuff, v3, p);

	vect = v3d((float)atof(v1), (float)atof(v2), (float)atof(v3));
	return vect;
}

float GrabFloat(char *linebuff, int p)
{
	char f[256];
	p = GetToken(linebuff, f, p);
	return (float)atof(f);
}



void CObjFile::ReadMtlFile(char *filename)
{
	std::ifstream myFile;
	char lineholder[1024];
	float tfloat;

	char directive[256];
	int p=0;
	int x=0;

	int curr_mtl = 0;

	SObjMaterialPtr NewMaterial;

	myFile.open(filename);
	if (myFile.is_open()) {
		std::cout << "Opened file " << filename << "\n";

		myFile.getline(lineholder, 1024);
		while (!myFile.eof()) {

			p = GetToken(lineholder, directive);
			if (p==-1) directive[0] = '\0';

			if (strcmp(directive, "newmtl") == 0) {
				// new material ... make, get name, add to list, select
				NewMaterial = new SObjMaterial;
				p = GetToken(lineholder, NewMaterial->sName, p);
				list_mtl.push_back(NewMaterial);
				curr_mtl = list_mtl.size()-1;

			} else if (strcmp(directive, "Kd") == 0) {
				// diffuse color
				list_mtl[curr_mtl]->diffuse = GrabVect(lineholder, p);

			} else if (strcmp(directive, "Ka") == 0) {
				// ambient color
				list_mtl[curr_mtl]->ambient = GrabVect(lineholder, p);

			} else if (strcmp(directive, "Ks") == 0) {
				// specular color
				list_mtl[curr_mtl]->specular = GrabVect(lineholder, p);

			} else if (strcmp(directive, "Ke") == 0) {
				// emissive color
				list_mtl[curr_mtl]->emissive = GrabVect(lineholder, p);

			} else if (strcmp(directive, "Ft") == 0) {
				// fresnel term
				list_mtl[curr_mtl]->fresnel_term = GrabFloat(lineholder, p);
			} else if (strcmp(directive, "d") == 0) {
				// alpha
				list_mtl[curr_mtl]->alpha = GrabFloat(lineholder, p);
			} else if (strcmp(directive, "illum") == 0) {
				tfloat = GrabFloat(lineholder, p);
				if (tfloat == 3) {
					list_mtl[curr_mtl]->shininess = list_mtl[curr_mtl]->specular*v3d(1,1,1)/3.0f;
				} else if (tfloat == 0) {
					list_mtl[curr_mtl]->lit = false;
				} else {
					list_mtl[curr_mtl]->shininess = 0.0;
				}
			} else if (strcmp(directive, "map_Kd")==0) {
				list_mtl[curr_mtl]->uses_texture = true;
				p = GetToken(lineholder, list_mtl[curr_mtl]->sTextureName, p);
			} else if (strcmp(directive, "map_Ka")==0) {
				list_mtl[curr_mtl]->uses_alpha_map = true;
				p = GetToken(lineholder, list_mtl[curr_mtl]->sAlphaMapName, p);
			} else {
			}
			myFile.getline(lineholder, 1024);
		}

		myFile.close();
	} else {
		std::cout << "Unable to open file.\n";
	}
}
