#pragma once
#include "VectorMath.h"

#include <iostream>
#include <vector>

struct SObjTri {
	unsigned int p1, p2, p3, tc1, tc2, tc3, n1, n2, n3;
	unsigned int mtl;
};

typedef SObjTri * SObjTriPtr;

struct SObjMaterial {
	SVect3d diffuse;
	SVect3d ambient;
	SVect3d specular;
	SVect3d emissive;
	float shininess;
	float index_refraction;
	float alpha;
	float fresnel_term;
	bool uses_texture;
	bool uses_alpha_map;
	bool lit;
	char sName[256];
	char sTextureName[1024];
	char sAlphaMapName[1024];

	SObjMaterial() {
		diffuse = v3d(0,0,0);
		ambient = v3d(0,0,0);
		specular = v3d(0,0,0);
		emissive = v3d(0,0,0);
		shininess = 0;
		index_refraction = 1;
		alpha = 1.0f;
		fresnel_term=0.0f;
		uses_texture=false;
		uses_alpha_map=false;
		lit=true;
	};
};

typedef SObjMaterial * SObjMaterialPtr;

class CObjGroup {
	public: 
		std::vector<SObjTriPtr> tri;
		unsigned int curr_mtl;
		char sName[1024];

		void AddTri(unsigned int p1, unsigned int p2, unsigned int p3, 
			unsigned int tc1, unsigned int tc2, unsigned int tc3, 
			unsigned int n1, unsigned int n2, unsigned int n3);
		void SetCurrentMtl(unsigned int mtl) { curr_mtl = mtl; };
		
		CObjGroup() {};
		~CObjGroup() {};
};

typedef CObjGroup * CObjGroupPtr;
typedef SVect3d * SVect3dPtr;

class CObjFile {
	public:
		std::vector<SVect3dPtr> list_v;
		std::vector<SVect3dPtr> list_vt;
		std::vector<SVect3dPtr> list_vn;
		std::vector<SObjMaterialPtr> list_mtl;
		std::vector<CObjGroupPtr> list_g;

		void ReadFile(const char * filename);
		void ReadMtlFile(char * filename);
		
		bool GetFilePath(char *outstr, const char *instr);
		int SetGroup(char *linebuff, int p);
		void ReadVertex(char *linebuff, int p);
		void ReadNormal(char *linebuff, int p);
		void ReadTexCoord(char *linebuff, int p);
		void ReadPolygon(char *linebuff, int p, unsigned int curr_mtl, unsigned int curr_group);

		CObjFile() {};
		~CObjFile() {};
};

int strpos(char *needle, char *haystack);
int GetToken(char *input, char *output, int startpos=0);