/*
Eos Camera - definitions
2013 - Matt Pierce

	Defines all the functions declared in eosCamera.h

TAB WIDTH: 4
*/
#include <math.h>
#include "VectorMath.h"
#include "eosCamera.h"

/// <summary>Pre-calculates some necessary values before you can use it
/// to transform screen coordinates to world coordinates</summary>
/// <returns>(void)</returns>
void eosPrepareCamera(eosCamera &camera) {
	// if we have valid screen width & height
	if (camera.w>0 && camera.h>0 && camera.fov>0 && camera.fov<3.1415926535897932384626433832795f) {
		camera.aspectRatio = (float)camera.w/(float)camera.h;
		camera.invTanTheta = 1.0f/tan(0.5f*camera.fov);
	}

	// now our position/aiming vectors
	camera.vecFwd = vmNormalize(v3d(camera.lookAt[0]-camera.p[0], camera.lookAt[1]-camera.p[1], camera.lookAt[2]-camera.p[2]));
	camera.vecUp = vmNormalize(v3d(camera.up[0], camera.up[1], camera.up[2]));
	camera.vecRight = vmNormalize(vmCrossProduct(camera.vecFwd, camera.vecUp));
	camera.vecUp = vmNormalize(vmCrossProduct(camera.vecRight, camera.vecFwd));
}

/// <summary>Converts screen coordinates to world coordinates. Supports fractional screen coordinates</summary>
/// <returns>(void)</returns>
void eosCameraProject(const eosCamera &camera, const float inX, const float inY, float *p, float *d) {
	float tx, ty, tz;

	tx = (2.0f*inX/(float)camera.w - 1.0f)*camera.aspectRatio;
	ty = (1.0f - 2.0f*inY/(float)camera.h);
	tz = camera.invTanTheta;

	d[0] = tx*camera.vecRight.x + ty*camera.vecUp.x + tz*camera.vecFwd.x;
	d[1] = tx*camera.vecRight.y + ty*camera.vecUp.y + tz*camera.vecFwd.y;
	d[2] = tx*camera.vecRight.z + ty*camera.vecUp.z + tz*camera.vecFwd.z;

	p[0] = camera.p[0];
	p[1] = camera.p[1];
	p[2] = camera.p[2];
}