/*
Eos Camera - header
2013 - Matt Pierce

	A tool to create a virtual camera to translate screen
	coordinates to world-space vectors ready to trace.

TAB WIDTH: 4
*/

#ifndef _EOS_CAMERA_H
#define _EOS_CAMERA_H

#include "VectorMath.h"

// eosCamera - A camera object for Eos
//		Not only does this describe the position and view of the
//		camera, it also describes the width and height of the
//		frameBuffer for our renderer. Make sure w/h are accurate
//		or the renderer will likely write to invalid memory and crash.
struct eosCamera {
	float			up[3];			// user-provided up vector
	float			p[3];			// user-provided position
	float			lookAt[3];		// user-provided point to look at
	unsigned int	w, h;			// user-provided screen width/height
	float			fov;			// user-provided vertical FOV (in radians)

	SVect3d			vecFwd;			// calculated forward vector
	SVect3d			vecRight;		// calculated right vector
	SVect3d			vecUp;			// calculated up vector
	float			invTanTheta;	// used in screen coord -> camera space
	float			aspectRatio;	// used in screen coord -> camera space
};

void eosPrepareCamera(eosCamera &camera);
void eosCameraProject(const eosCamera &camera, const float inX, const float inY, float *p, float *d);
#endif // _EOS_CAMERA_H