/*
Eos Light - definitions
2013 - Matt Pierce

	Defines functions declared in eosLight.h

TAB WIDTH: 4
*/
#include "eosLight.h"

void eosPrepareLight(eosLight *l) {
	if ((l->type == EOS_TYPE_POINT) && (l->falloff == EOS_FALLOFF_ORB_LIGHT)) {
		// calculate the effective radius of an orb light
		float i = (l->col[0] + l->col[1] + l->col[2])/3.0f;
		l->r = (i*l->fov -  0.0001f*l->fov)/(0.0001f);
		l->r = sqrtf(l->r);
	}

	if (l->r > 0.0f) {
		l->r2 = l->r*l->r;
		l->box.min[0] = l->p[0] - l->r;
		l->box.min[1] = l->p[1] - l->r;
		l->box.min[2] = l->p[2] - l->r;
		l->box.max[0] = l->p[0] + l->r;
		l->box.max[1] = l->p[1] + l->r;
		l->box.max[2] = l->p[2] + l->r;
	} else if (l->r<0.0f) {
		l->r = 0.0f;
	}

	if (l->type == EOS_TYPE_SPOTLIGHT) {
		l->cosFov = (float)cos(l->fov);
	}

	if (l->type == EOS_TYPE_DIRECTIONAL_SOFT) {
		CMatrix		tMatrix;
		SVect3d		d2, d3, d4;
		float		theta, phi, r;

		d2 = -vmNormalize(v3d(l->d[0], l->d[1], l->d[2]));

		if (d2 == v3d(0, 1, 0)) {
			tMatrix.LoadIdentity();
			tMatrix.Rotate(1.5707963267948966192313216916398f, 1, 0, 0);
		} else if (d2 == v3d(0, 1, 0)) {
			tMatrix.LoadIdentity();
			tMatrix.Rotate(-1.5707963267948966192313216916398f, 1, 0, 0);
		} else {
			d3 = d2;
			d3.y = 0.0f;
			d3 = vmNormalize(d3);
			phi = vmGetAngle(d3, d2);
			if (d2.y<0.0f) phi = -phi;
			theta = vmGetAngle(d3, v3d(0.0f, 0.0f, -1.0f));
			if (d3.x>0.0f) theta = -theta;
			tMatrix.LoadIdentity();
			tMatrix.Rotate(phi, 1.0f, 0.0f, 0.0f);
			tMatrix.Rotate(theta, 0.0f, 1.0f, 0.0f);
		}

		r = tan(l->fov/2.0f);
		d2 = v3d(0.0f, r*sin(1.5707963267948966192313216916398f), -1.0f);
		d3 = v3d(r*cos(3.6651914291880921115397506138261f), r*sin(3.6651914291880921115397506138261f), -1.0f);
		d4 = v3d(r*cos(5.7595865315812876038481795360124f), r*sin(5.7595865315812876038481795360124f), -1.0f);

		d2 = tMatrix.Transform(d2);
		d3 = tMatrix.Transform(d3);
		d4 = tMatrix.Transform(d4);

		l->d2[0] = d2.x;
		l->d2[1] = d2.y;
		l->d2[2] = d2.z;

		l->d3[0] = d3.x;
		l->d3[1] = d3.y;
		l->d3[2] = d3.z;

		l->d4[0] = d4.x;
		l->d4[1] = d4.y;
		l->d4[2] = d4.z;
	}
}