/*
Eos Light - header
2013 - Matt Pierce

	Defines a light source and its associated functions

	__declspec(dllexport) employed to see if I can make a
	DLL out of this later

TAB WIDTH: 4
*/

#ifndef _EOS_LIGHT_H
#define _EOS_LIGHT_H

#include "eosTree.h"			// for eosAABB

enum _declspec(dllexport) eosLightType {
	EOS_TYPE_AMBIENT,
	EOS_TYPE_POINT,
	EOS_TYPE_DIRECTIONAL,
	EOS_TYPE_SPOTLIGHT,
	EOS_TYPE_GRADIENT,
	EOS_TYPE_DIRECTIONAL_SOFT
};

enum _declspec(dllexport) eosLightFalloff {
	EOS_FALLOFF_LINEAR,
	EOS_FALLOFF_INVERSE,
	EOS_FALLOFF_INVERSE_SQUARED,
	EOS_FALLOFF_ORB_LIGHT				// should use infinite range, eosLight.fov represents radius of a spherical light orb, squared
};

// eosLight
//		Defines a light that can be added to a scene. Not
//		all of these variables are used for every type of
//		light, but since there are only a handful of lights
//		in the scene anyway we're looking at only a few
//		wasted bytes in total.
struct _declspec(dllexport) eosLight {
	float			p[3];			// light position
	float			d[3];			// light direction (optional)
	float			r;				// light radius (0 = unlimited)
	float			r2;				// radius squared
	float			fov;			// field of view (for spotlights and directional soft lights)
	float			cosFov;			// cosine of fov (so we can check spotlight in/out with simple dot product)
	float			col[3];			// light color
	float			col2[3];		// secondary color (for gradient type)
	eosLightType	type;			// light type
	eosLightFalloff	falloff;		// light falloff routine
	eosAABB			box;			// a bounding box for ranged lights
	float			d2[3];			// for directional soft
	float			d3[3];			// for directional soft
	float			d4[3];			// for directional soft
};

void eosPrepareLight(eosLight *l);

#endif //_EOS_LIGHT_H