/*
Eos Renderer - header
2013 - Matt Pierce

	Provides a function to call the renderer, along with a few options.

TAB WIDTH: 4
*/


#ifndef _EOS_RENDERER_H
#define _EOS_RENDERER_H

#include "eosCamera.h"
#include "eosScene.h"

// define our default values
const int		EOS_RENDERER_LEVELS = 6;						// number of levels of reflections
const int		EOS_RENDERER_THREADS = 20;						// number of threads to spawn
const float		EOS_RENDERER_SUBDIVISION_THRESHOLD = 0.25f;		// the threshold number for subsample rendering
const int		EOS_RENDERER_ANTIALIAS_LEVELS = 3;				// the number of times to recurse for antialiasing
const float		EOS_RENDERER_ANTIALIAS_THRESHOLD = 0.25f;		// how different the next pixel has to be to recurse
const float		EOS_RENDERER_SOFT_LIGHT_THRESHOLD = 0.25f;		// when to subdivide soft lights

// - FLAG VALUES -
// The pre-processor will substitute these in, and then the
// optimizer will replace them with literals - should result
// in zero execution time.

// LOWER 4 BITS - type of renderer
#define EOS_RENDERER_STANDARD				(1 << 0)
#define EOS_RENDERER_SUBSAMPLE				(1 << 1)
#define EOS_RENDERER_MULTITHREADED			(1 << 2)

// UPPER 12 BITS - quality settings
#define EOS_RENDERER_ANTIALIAS				(1 << 4)
#define EOS_RENDERER_SUBSAMPLE_DIFFUSE		(1 << 5)
#define EOS_RENDERER_SUBSAMPLE_SPECULAR		(1 << 6)
#define EOS_RENDERER_SUBSAMPLE_REFLECTION	(1 << 7)
#define EOS_RENDERER_AMBIENT_OCCLUSION		(1 << 8)
#define EOS_RENDERER_SUBSAMPLE_DEBUG		(1 << 9)
#define EOS_RENDERER_JITTER_SOFT_SHADOWS	(1 << 10)

class eosRendererSettings {
	// I would normally make this a struct to avoid class overhead, but
	// I really want a constructor on this.
	public:
		// variables
		int				levels;
		int				threads;
		float			subdivisionThreshold;
		float			reflectionThreshold;
		int				antialiasLevels;
		float			antialiasThreshold;
		unsigned int	flags;
		float			fogRate;
		float			fogColor[3];
		float			ambientOcclusionDistance;
		float			ambientOcclusionSamples;
		float			softLightThreshold;

		// functions
		eosRendererSettings():
			levels(EOS_RENDERER_LEVELS),
			threads(EOS_RENDERER_THREADS),
			subdivisionThreshold(EOS_RENDERER_SUBDIVISION_THRESHOLD),
			reflectionThreshold(EOS_RENDERER_SUBDIVISION_THRESHOLD),
			antialiasLevels(EOS_RENDERER_ANTIALIAS_LEVELS),
			antialiasThreshold(EOS_RENDERER_ANTIALIAS_THRESHOLD),
			flags(EOS_RENDERER_STANDARD),
			fogRate(0),
			ambientOcclusionDistance(-1),
			ambientOcclusionSamples(10),
			softLightThreshold(EOS_RENDERER_SOFT_LIGHT_THRESHOLD) {};

};

void eosRenderScene(eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings);

#endif // _EOS_RENDERER_H
