/*
Eos Renderer Mesh - header
2013 - Matt Pierce

	Links a B-KD Tree and its triangles with information
	used by the renderer. Contains its own matrix so that
	you can instance the same B-KD tree multiple times.

TAB WIDTH: 4
*/


#ifndef _EOS_RENDERER_MESH_H
#define _EOS_RENDERER_MESH_H

#include "VectorMath.h"
#include "eosTriangle.h"
#include "eosTriangleRenderInfo.h"
#include "eosMaterial.h"
#include "eosTree.h"

#include "ObjLoader.h"

// eosRendererMesh
//		A complete mesh for the renderer. Geometry is stored in
//		tris, rendering info for each tri in renderInfo, a
//		quick reference to materials, the tree, and a matrix.
//		Because everything's stored by pointer, you can use this
//		to instance another already-existing object and move it
//		around with the mesh's matrix.
struct eosRendererMesh {
	unsigned int			nTris;			// number of triangles
	unsigned int			nMats;			// number of materials
	eosTriangle *			tris;			// triangles to trace	
	eosTriangleRenderInfo *	renderInfo;		// Information about each triangle used by the renderer
	eosMaterial *			mats;			// array of materials
	eosTree *				tree;			// B-KD tree of this.tris
	CMatrix					matrix;			// our own transformation matrix
};

eosTraceResult eosTraceMesh(eosRendererMesh *mesh,
							const float &px, const float &py, const float &pz,
							const float &dx, const float &dy, const float &dz,
							float capT=-1.0f);

void eosRendererMeshFromObj (const char *inFilename, eosRendererMesh* inMesh);

#endif // _EOS_RENDERER_MESH_H