/*
Eos Scene - header
2013 - Matt Pierce

	Collects information to describe a complete scene:
	Meshes, Lights, and Environment Map

	A NOTE ON HOW SCENES ARE INTENDED TO WORK:
	The eosScene struct only stores pointers to objects. The
	caller is responsible for allocating/deallocating the
	objects whenever necessary. It's recommended that you
	store your objects in ways that make sense for your game,
	clear the scene at the beginning of each frame, and then
	add the things you want to render every time. Because we're
	talking about (likely) < 100 pointers here there's very
	little impact on overall performance at all.

TAB WIDTH: 4
*/

#ifndef _EOS_SCENE_H
#define _EOS_SCENE_H

#include "eosRendererMesh.h"
#include "eosEnvironmentMap.h"
#include "eosLight.h"

// Determines how frequently the Add functions
// should result in allocating more memory.
// Low numbers mean lower memory overhead.
// High numbers mean less-frequent allocations w/ copy.
const int EOS_SCENE_ALLOCATION_INCREMENT = 50;

// eosScene
//		Collects a bunch of information needed to
//		describe a scene. This, combined with a
//		camera to describe your view, should be
//		all you need to send to the renderer.
struct eosScene {
	int							nLights, capLights;		// number and capacity of lights
	eosLight **					lights;					// array of ptrs to lights
	int							nMeshes, capMeshes;		// number and capacity of meshes
	eosRendererMesh **			meshes;					// array of ptrs to meshes
	eosEnvironmentMap *			env;					// ptr to env map
};

eosScene		eosInitScene();
void			eosAddLight(eosScene *scene, eosLight *light);
void			eosClearLights(eosScene *scene);
void			eosAddMesh(eosScene *scene, eosRendererMesh *mesh);
void			eosClearMeshes(eosScene *scene);
void			eosAddEnvMap(eosScene *scene, eosEnvironmentMap *env);
void			eosClearEnvMap(eosScene *scene);
void			eosClearScene(eosScene *scene);

eosTraceResult	eosTraceScene(eosScene *scene, 
								const float &px, const float &py, const float &pz, 
								const float &dx, const float &dy, const float &dz, 
								const float &capT = -1.0f);

#endif // _EOS_SCENE_H