/*
Eos Texture - header
2013 - Matt Pierce

	Outlines a texture object for Eos similar to OpenGL
	that holds some properties and can be sampled
	by the renderer.

TAB WIDTH: 4
*/

#ifndef _EOS_TEXTURE_H
#define _EOS_TEXTURE_H

// texture repeat options
#define EOS_TEXTURE_CLAMP 0x01
#define EOS_TEXTURE_REPEAT 0x02

// texture data formats
#define EOS_TEXTURE_BYTE 0x01
#define EOS_TEXTURE_FLOAT 0x02

// texture filter options
#define EOS_TEXTURE_NEAREST 0x01
#define EOS_TEXTURE_BILINEAR 0x02

// eosTexture
//		Describes a texture (duh) with a buffer,
//		width & height, and associated settings
struct eosTexture {
	int				w, h;		// width and height (must be a power of 2)
	void *			buff;		// pointer to data
	unsigned char	planes;		// number of planes (1-4)
	unsigned char	type;		// EOS_TEXTURE_BYTE, EOS_TEXTURE_FLOAT
	unsigned char	filter;		// EOS_TEXTURE_NEAREST, EOS_TEXTURE_BILINEAR
	unsigned char	repeatX;	// EOS_TEXTURE_CLAMP, EOS_TEXTURE_REPEAT
	unsigned char	repeatY;	// EOS_TEXTURE_CLAMP, EOS_TEXTURE_REPEAT
};

eosTexture			eosCreateTexture(int width, int height, unsigned char planes, 
									unsigned char type, unsigned char filter, 
									unsigned char repeatX, unsigned char repeatY, 
									void * buff);
void				eosSampleTexture(const eosTexture &tex, float s, float t, float *output);

#endif //_EOS_TEXTURE_H