/*
Eos Triangle - header
2013 - Matt Pierce

	Defines a triangle in the raytracer, and a structure
	to store the result of tracing a ray toward a triangle.

	__declspec(dllexport) employed to see if I can make a
	DLL out of this later

TAB WIDTH: 4
*/


#ifndef _EOS_TRIANGLE_H
#define _EOS_TRIANGLE_H

// eosTraceResult
//		This is what's returned when you trace a triangle.
struct __declspec(dllexport) eosTraceResult {
	float			t;				// the point along the ray at which the intersection occured
									// (negative value indicates no contact)
	unsigned int	tri;			// the index of the triangle that was contacted
									// (refers to an array in the kD tree - kD tree responsible for filling this out)
	unsigned int	mesh;			// the index of the mesh that the triangle's in
									// (refers to an array in the eosScene - the eosScene is responsible for filling this out)
	float			a1, a2, a3;		// the point on the triangle at which the intersection occured
									// (point calculated as (p1*a1 + p2*a2 + p3*a3))
};

// eosTriangle
//		Describes a single triangle as understood by the
//		raytracer. This is ONLY the part of the triangle
//		requred to trace the scene. Information needed
//		to render the triangle (textures, normals, etc ...)
//		should be handled by the rendering engine.
struct __declspec(dllexport) eosTriangle {
	float			p1[3], p2[3], p3[3];	// the three vertices that make the triangle
	float			pnx, pny, pnz, pd;		// plane equation for plane described by vertices
	float			bu,bv,cu,cv;			// some values used frequently while checking intersections
											// (these get precalcualted to speed intersetion checks)
	unsigned char	PrimaryAxis;			// the axis the plane most closely faces
											// (used to calculate (a1, a2, a3) on intersect)
};

void __declspec(dllexport) eosSetupTriangle(eosTriangle &tri);
eosTraceResult __declspec(dllexport) eosTestTrace(const eosTriangle &tri, 
	const float &px, const float &py, const float &pz, 
	const float &dx, const float &dy, const float &dz);

#endif // _EOS_TRIANGLE_H