#include <stdio.h>
#include <time.h>
#include "sampleScene.h"

#define STB_IMAGE_RESIZE_IMPLEMENTATION 1
#include "../stb_master/stb_image_resize.h"

void				GSSampleScene::SetFullScreen(bool inFullScreen) {
	if (this->GetFullScreen()) {
		this->FullScreenWidth = this->myInitData.screen->w;
		this->FullScreenHeight = this->myInitData.screen->h;
	}
	this->myInitData.FullScreen = inFullScreen;

	SDL_FreeSurface(this->vscrSurface);
	free(this->vscr);

	if (inFullScreen == true ) {
		this->myInitData.screen = SDL_SetVideoMode(this->FullScreenWidth, this->FullScreenHeight, 32, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_FULLSCREEN);
	} else {
		this->myInitData.screen = SDL_SetVideoMode(this->myInitData.DefaultScreenWidth, this->myInitData.DefaultScreenHeight, 32, SDL_HWSURFACE | SDL_DOUBLEBUF);
	}


	this->vscr = (unsigned char *)malloc(this->myInitData.screen->w*this->myInitData.screen->h*3*sizeof(unsigned char));
	this->vscrSurface = SDL_CreateRGBSurfaceFrom(this->vscr, this->myInitData.screen->w, this->myInitData.screen->h, 24, this->myInitData.screen->w*3, 0x0000FF, 0x00FF00, 0xFF0000, 0);
}

void				GSSampleScene::Init(void *initData) {

	// copy initData
	SampleSceneInitData *	tmpData = (SampleSceneInitData *)initData;
	this->myInitData = *tmpData;	// switch to item-by-item if we add anything dynamically allocated

	// init player data
	this->PlayerPos = v3d(0.0f, 0.0f, 0.0f);
	this->PlayerHeading = 0.0f;
	this->PlayerPitch = 0.0f;

	// init camera
	this->testCamera.w = this->myInitData.InternalWidth;
	this->testCamera.h = this->myInitData.InternalHeight;
	this->testCamera.fov = 3.14159f/2.0f;
	this->testCamera.up[0] = 0.0f;
	this->testCamera.up[1] = 1.0f;
	this->testCamera.up[0] = 0.0f;

	// init exposure tracker
	eosInitExposureTracker(&(this->tracker));
	this->tracker.minExposure = 1.0f;
	this->tracker.maxExposure = 10.0f;

	// init animation
	this->AnimTimer = 0.0;

	// init display settings
	this->currentRS = 0;
	this->currentCamera = 0;
	this->subsampleDebug = false;

	// init virtual screen
	if (this->vscr == NULL) {
		this->vscr = (unsigned char *)malloc(this->myInitData.screen->w*this->myInitData.screen->h*3*sizeof(unsigned char));
		this->vscrSurface = SDL_CreateRGBSurfaceFrom(this->vscr, this->myInitData.screen->w, this->myInitData.screen->h, 24, this->myInitData.screen->w*3, 0x0000FF, 0x00FF00, 0xFF0000, 0);
	}

}

void				GSSampleScene::Load(void) {

	// load checkerboard texture
	this->checkerBoard = this->LoadTextureFromSDL(".\\checker-board.png");
	this->checkerBoard.filter = EOS_TEXTURE_NEAREST;
	
	// load sand texture
	this->sand = this->LoadTextureFromSDL(".\\sand.jpg");
	this->sand.filter = EOS_TEXTURE_BILINEAR;

	// load env map
	int w=0, h=0;
	unsigned char *buff = NULL;
	float * fbuff=NULL;
	fbuff = this->LoadRawFloat("xpos.float", 256, 256, true);
	if (fbuff) {
		this->env.xpos = (eosTexture *)malloc(sizeof(eosTexture));
		*(this->env.xpos) = eosCreateTexture(256, 256, 3, EOS_TEXTURE_FLOAT, EOS_TEXTURE_BILINEAR, EOS_TEXTURE_CLAMP, EOS_TEXTURE_CLAMP, fbuff);
	}
	fbuff = this->LoadRawFloat("xneg.float", 256, 256, true);
	if (fbuff) {
		this->env.xneg = (eosTexture *)malloc(sizeof(eosTexture));
		*(this->env.xneg) = eosCreateTexture(256, 256, 3, EOS_TEXTURE_FLOAT, EOS_TEXTURE_BILINEAR, EOS_TEXTURE_CLAMP, EOS_TEXTURE_CLAMP, fbuff);
	}
	fbuff = this->LoadRawFloat("ypos.float", 256, 256, true);
	if (fbuff) {
		this->env.ypos = (eosTexture *)malloc(sizeof(eosTexture));
		*(this->env.ypos) = eosCreateTexture(256, 256, 3, EOS_TEXTURE_FLOAT, EOS_TEXTURE_BILINEAR, EOS_TEXTURE_CLAMP, EOS_TEXTURE_CLAMP, fbuff);
	}
	fbuff = this->LoadRawFloat("yneg.float", 256, 256, true);
	if (fbuff) {
		this->env.yneg = (eosTexture *)malloc(sizeof(eosTexture));
		*(this->env.yneg) = eosCreateTexture(256, 256, 3, EOS_TEXTURE_FLOAT, EOS_TEXTURE_BILINEAR, EOS_TEXTURE_CLAMP, EOS_TEXTURE_CLAMP, fbuff);
	}
	fbuff = this->LoadRawFloat("zpos.float", 256, 256, true);
	if (fbuff) {
		this->env.zpos = (eosTexture *)malloc(sizeof(eosTexture));
		*(this->env.zpos) = eosCreateTexture(256, 256, 3, EOS_TEXTURE_FLOAT, EOS_TEXTURE_BILINEAR, EOS_TEXTURE_CLAMP, EOS_TEXTURE_CLAMP, fbuff);
	}
	fbuff = this->LoadRawFloat("zneg.float", 256, 256, true);
	if (fbuff) {
		this->env.zneg = (eosTexture *)malloc(sizeof(eosTexture));
		*(this->env.zneg) = eosCreateTexture(256, 256, 3, EOS_TEXTURE_FLOAT, EOS_TEXTURE_BILINEAR, EOS_TEXTURE_CLAMP, EOS_TEXTURE_CLAMP, fbuff);
	}

	// load font
	this->myFont = TTF_OpenFont("Prelude-Medium.ttf", 12);
	TTF_SetFontHinting(this->myFont, TTF_HINTING_LIGHT);			// as per recommendation from SDL_ttf site

	// load meshes - using factories in sample, but would be from files in game
	this->BuildTorus(this->torusMesh, 0.8f, 0.2f, 300, 150);
	this->torusMesh2 = this->torusMesh; // trying copying all the pointers
										// no addl storage, relying on mesh's matrix
	this->GenerateHeightmap();
	this->heightMap.mats[0].tex = &(this->sand);

	char *filePrepend = ".\\";
	char fileName[1024] = "";
	sprintf(fileName, "%s%s", filePrepend, "venus.obj");
	eosRendererMeshFromObj(fileName, &this->loadedMesh);
	for (int matIndex = 0; matIndex<this->loadedMesh.nMats; matIndex++) {
		if (this->loadedMesh.mats[matIndex].texFile) {
			this->loadedMesh.mats[matIndex].tex = (eosTexture *)malloc(sizeof(eosTexture));
			sprintf(fileName, "%s%s", filePrepend, this->loadedMesh.mats[matIndex].texFile);
			*(this->loadedMesh.mats[matIndex].tex) = this->LoadTextureFromSDL(fileName);
			this->loadedMesh.mats[matIndex].tex->filter = EOS_TEXTURE_BILINEAR;
			free(this->loadedMesh.mats[matIndex].texFile);
		}
	}

	// load lights - would be part of some scene description
	this->tmpLight = (eosLight *)malloc(sizeof(eosLight));
	this->tmpLight->type = EOS_TYPE_DIRECTIONAL_SOFT;
	this->tmpLight->d[0] = -0.375f;
	this->tmpLight->d[1] = -0.484f;
	this->tmpLight->d[2] = -1.0f;
	this->tmpLight->col[0] = 0.75f; 
	this->tmpLight->col[1] = 0.70f;
	this->tmpLight->col[2] = 0.65f;
	this->tmpLight->r = 0.0f;
	this->tmpLight->fov = 2.0f*0.01745329251994329576923690768489f; // the sun is about 1/2 degree across in a skydome

	this->ambientLight = (eosLight *)malloc(sizeof(eosLight));
	this->ambientLight->type = EOS_TYPE_GRADIENT;
	this->ambientLight->col[0] = 0.48f*0.20f;
	this->ambientLight->col[1] = 0.56f*0.20f;
	this->ambientLight->col[2] = 0.68f*0.20f;
	this->ambientLight->col2[0] = 0.03f;
	this->ambientLight->col2[1] = 0.03f;
	this->ambientLight->col2[2] = 0.03f;
	this->ambientLight->d[0] = 0.0f;
	this->ambientLight->d[1] = 1.0f;
	this->ambientLight->d[0] = 0.0f;

	this->localLight = (eosLight *)malloc(sizeof(eosLight));
	this->localLight->type = EOS_TYPE_POINT;
	this->localLight->col[0] = 10.0f;
	this->localLight->col[1] = 10.0f;
	this->localLight->col[2] = 10.0f;
	this->localLight->p[0] = 0.0f;
	this->localLight->p[1] = 10.0f;
	this->localLight->p[2] = 0.0f;
	this->localLight->fov = 3.0f;
	this->localLight->fov *= this->localLight->fov;
	this->localLight->falloff = EOS_FALLOFF_ORB_LIGHT;

	// create scene
	this->testScene = eosInitScene();
	// eosAddMesh(&this->testScene, &this->torusMesh);			// add meshes
	// eosAddMesh(&this->testScene, &this->torusMesh2);
	eosAddMesh(&this->testScene, &this->heightMap);
	eosAddMesh(&this->testScene, &this->loadedMesh);
	eosAddLight(&this->testScene, this->tmpLight);			// sunlight
	eosAddLight(&this->testScene, this->ambientLight);		// ambient light
	// eosAddLight(&this->testScene, this->localLight);		// local light (with falloff)
	eosAddEnvMap(&this->testScene, &this->env);				// add environment map

	// set loaded flag
	this->IsLoaded = true;

}

void				GSSampleScene::Unload(void) {

	if (this->IsLoaded) {
		// free font
		if (this->myFont) TTF_CloseFont(this->myFont);
		this->myFont = NULL;

		// free environment map
		free(this->env.xneg->buff);
		free(this->env.xpos->buff);
		free(this->env.yneg->buff);
		free(this->env.ypos->buff);
		free(this->env.zneg->buff);
		free(this->env.zpos->buff);

		// free sand texture
		free(this->sand.buff);

		// free checkerboard texture
		free(this->checkerBoard.buff);

		// free meshes
		free(this->torusMesh.mats);
		free(this->torusMesh.tris);
		free(this->torusMesh.renderInfo);
		eosClearTree(this->torusMesh.tree);
		free(this->torusMesh.tree);

		free(this->heightMap.mats);
		free(this->heightMap.tris);
		free(this->heightMap.renderInfo);
		eosClearTree(this->heightMap.tree);
		free(this->heightMap.tree);

		// free lights
		free(this->tmpLight);
		free(this->ambientLight);
		free(this->localLight);

		this->IsLoaded = false;
	}
}

void				GSSampleScene::Cycle(float dt, void *cycleData) {
	SDL_Event		event;
	SVect3d			PlayerFVect, PlayerRVect;
	CMatrix			PlayerMatrix;
	Uint8 *			keys;

	time_t			RawTime;
	struct tm *		TimeInfo;
	char			OutBMPName[256];

	PlayerMatrix.LoadIdentity();
	PlayerMatrix.Rotate(this->PlayerHeading, 0.0f, 1.0f, 0.0f);
	PlayerFVect = PlayerMatrix.Transform(v3d(0.0f, 0.0f, -1.0f));
	PlayerRVect = PlayerMatrix.Transform(v3d(1.0f, 0.0f, 0.0f));

	keys = SDL_GetKeyState(NULL);
	if (keys['w']) this->PlayerPos = this->PlayerPos + PlayerFVect*(this->myInitData.PlayerKeyboardSensitivity)*dt*1000.0f;
	if (keys['s']) this->PlayerPos = this->PlayerPos - PlayerFVect*(this->myInitData.PlayerKeyboardSensitivity)*dt*1000.0f;
	if (keys['a']) this->PlayerPos = this->PlayerPos - PlayerRVect*(this->myInitData.PlayerKeyboardSensitivity)*dt*1000.0f;
	if (keys['d']) this->PlayerPos = this->PlayerPos + PlayerRVect*(this->myInitData.PlayerKeyboardSensitivity)*dt*1000.0f;



	while (SDL_PollEvent(&event)) {
		if (event.type == SDL_QUIT) {
			// find some way to tell the system to quit completely
			*(this->myInitData.gameRunning) = 0;
		}
		if (event.type == SDL_KEYDOWN) {
		}
		if (event.type == SDL_KEYDOWN) {
			if (event.key.keysym.sym<256) keys[event.key.keysym.sym] = 1;
			switch(event.key.keysym.sym) {
			case SDLK_RIGHT:
				this->currentRS = (this->currentRS+1)%(this->nRendererStates);
				break;
			case SDLK_LEFT:
				this->currentRS = (this->nRendererStates + this->currentRS - 1)%(this->nRendererStates);
				break;
			case SDLK_ESCAPE:
				// find some way to tell the system to unload me
				*(this->myInitData.gameRunning) = 0;
				break;
			case 'z':
				this->ambientOcclusionDebug = !this->ambientOcclusionDebug;
				break;
			case 'x':
				this->subsampleDebug = !this->subsampleDebug;
				break;
			case 'c':
				this->currentCamera = (this->currentCamera+1)%(this->nCameras);
				break;
			case 'i':
				this->Init((void *)(&this->myInitData));
				break;
			case 'o':
				this->Unload();
				this->Load();
				break;
			case 'f':
				this->SetFullScreen(!this->GetFullScreen());
				break;
			case 'r':
				this->outputStats = !this->outputStats;
				break;
			case SDLK_F12:
				RawTime = time(NULL);
				TimeInfo = localtime(&RawTime);
				sprintf(OutBMPName, "%04d-%02d-%02d_eos_screenshot_%d.bmp", TimeInfo->tm_year+1900, TimeInfo->tm_mon+1, TimeInfo->tm_mday, RawTime);
				SDL_SaveBMP(this->vscrSurface, OutBMPName);
				break;
			default:
				break;
			}
		}
		if (event.type == SDL_MOUSEBUTTONDOWN) {
		}
		if (event.type == SDL_MOUSEBUTTONUP) {
		}
		if (event.type == SDL_MOUSEMOTION) {
			this->PlayerHeading -= event.motion.xrel*this->myInitData.PlayerMouseSensitivity;
			this->PlayerPitch += event.motion.yrel*this->myInitData.PlayerMouseSensitivity;
			if (this->PlayerPitch > 1.5) this->PlayerPitch = 1.5;
			if (this->PlayerPitch < -1.5) this->PlayerPitch = -1.5;
		}
	}

	// this is just a sample scene, so we just advance an animation timer here. If it were a game
	// we'd use dt to figure out some physics/simulation
	this->dt = dt;
	this->AnimTimer += dt*60;
}

void				GSSampleScene::Draw() {


	int i, j;
	unsigned char * vscrTemp = NULL;
	const float pi = 3.1415926535897932384626433832795f;

	float *frameBuffer;		// frameBuffer
	float *frameBufferTemp;	// pointer to frameBuffer

	char FrameTitle[256];

	// clear screen
	memset(this->vscr, 0, this->myInitData.InternalWidth*this->myInitData.InternalHeight*3);


	// rings going through each other  
	this->torusMesh.matrix.LoadIdentity();
	this->torusMesh.matrix.Rotate(this->AnimTimer*0.01f+pi, 0, 0, 1);
	this->torusMesh.matrix.Scale(2.0f + cos(this->AnimTimer*0.01f), 2.0f + cos(this->AnimTimer*0.01f), 2.0f + cos(this->AnimTimer*0.01f));
	this->torusMesh.matrix.Translate(0.0f, 2*sin(this->AnimTimer*0.01f), 0.0f);

	this->torusMesh2.matrix.LoadIdentity();
	this->torusMesh2.matrix.Rotate(this->AnimTimer*0.01f, 1, 0, 0);
	this->torusMesh2.matrix.Scale(2.0f + cos(this->AnimTimer*0.01f+pi), 2.0f + cos(this->AnimTimer*0.01f+pi), 2.0f + cos(this->AnimTimer*0.01f+pi));
	this->torusMesh2.matrix.Translate(0.0f, 2*sin(this->AnimTimer*0.01f+pi), 0.0f);

	// our loaded mesh
	if (this->currentCamera == 3) {
		this->loadedMesh.matrix.LoadIdentity();
		this->loadedMesh.matrix.Scale(3.0f, 3.0f, 3.0f);
		// this->loadedMesh.matrix.Rotate(this->AnimTimer*0.01f, 0, 1, 0);
		this->loadedMesh.matrix.Translate(0.0f, 5.0f, -0.0f);
	} else {
		this->loadedMesh.matrix.LoadIdentity();
		this->loadedMesh.matrix.Scale(1.5f, 1.5f, 1.5f);
		this->loadedMesh.matrix.Rotate(this->AnimTimer*0.01f, 0, 1, 0);
		this->loadedMesh.matrix.Translate(0.0f, 0.0f, -0.0f);
	}

	// position Heightmap
	this->heightMap.matrix.LoadIdentity();
	this->heightMap.matrix.Scale(40.0f, 20.0f, 40.0f);
	this->heightMap.matrix.Translate(0, 205, 0);

	// position camera
	float t = this->AnimTimer * 0.001f;
	SVect3d temp = v3d(0,0,0);
	CMatrix tMat;
	eosTraceResult res;
	this->testCamera.up[0] = 0.0f;
	this->testCamera.up[1] = 1.0f;
	this->testCamera.up[2] = 0.0f;
	this->testCamera.fov = 1.5707963267948966192313216916398f;
	switch (this->currentCamera) {
		case 0:
			// circle the center
			this->testCamera.lookAt[0] = 0.0f;
			this->testCamera.lookAt[1] = 0.0f;
			this->testCamera.lookAt[2] = 0.0f;
			this->testCamera.p[0] = 3*sin(this->AnimTimer*0.0017f);
			this->testCamera.p[1] = 2*sin(this->AnimTimer*0.0013f);
			this->testCamera.p[2] = 3*cos(this->AnimTimer*0.0017f);
			break;
		case 1:
			this->testCamera.p[0] = (100.0f*sin(2*t)) * cos(t);
			this->testCamera.p[1] = 30.0f * (1.0f -cos(4*t)) - 2.0f;
			this->testCamera.p[2] = (100.0f*sin(2*t)) * sin(t);
			t += 0.001f;
			this->testCamera.lookAt[0] = (100.0f*sin(2*t)) * cos(t);
			this->testCamera.lookAt[1] = 30.0f * (1.0f - cos(4*t)) - 2.0f;
			this->testCamera.lookAt[2] = (100.0f*sin(2*t)) * sin(t);
			t += 0.01f;
			temp.x = (100.0f*sin(2*t)) * cos(t);
			temp.y = 30.0f * (1.0f - cos(4*t)) - 2.0f;
			temp.z = (100.0f*sin(2*t)) * sin(t);
			this->torusMesh.matrix.LoadIdentity();
			this->torusMesh.matrix.Rotate(1.57f, 1, 0, 0);
			this->torusMesh.matrix.Translate(temp.x, temp.y, temp.z);
			this->torusMesh2.matrix.LoadIdentity();
			this->torusMesh2.matrix.Translate(temp.x, temp.y, temp.z);
			break;
		case 2:
			this->testCamera.fov = 1.5707963267948966192313216916398f/6.0f;
			temp.x = (100.0f*sin(2*t)) * cos(t);
			temp.y = 30.0f * (1.0f - cos(4*t));
			temp.z = (100.0f*sin(2*t)) * sin(t);
			this->torusMesh.matrix.LoadIdentity();
			this->torusMesh.matrix.Rotate(1.57f, 1, 0, 0);
			this->torusMesh.matrix.Translate(temp.x, temp.y, temp.z);
			this->torusMesh2.matrix.LoadIdentity();
			this->torusMesh2.matrix.Translate(temp.x, temp.y, temp.z);
			this->testCamera.p[0] = temp.x - 8.0f;
			this->testCamera.p[1] = temp.y + 1.0f;
			this->testCamera.p[2] = temp.z + 4.0f;
			this->testCamera.lookAt[0] = temp.x;
			this->testCamera.lookAt[1] = temp.y;
			this->testCamera.lookAt[2] = temp.z;
			break;
		case 3:
			res = eosTraceMesh(&(this->heightMap), this->PlayerPos.x, 10000.0f, this->PlayerPos.z, 0.0f, -1.0f, 0.0f, -1.0f);
			if (res.t<0.0f) {
				this->PlayerPos.y = 0.0f;
			} else {
				this->PlayerPos.y = 10000.0f - res.t + 10.0f;
			}
			tMat.LoadIdentity();
			tMat.Rotate(this->PlayerPitch, 1.0f, 0.0f, 0.0f);
			tMat.Rotate(this->PlayerHeading, 0.0f, 1.0f, 0.0f);
			tMat.Translate(this->PlayerPos.x, this->PlayerPos.y, this->PlayerPos.z);
			temp = tMat.Transform(v3d(0,0,-1));
			this->testCamera.p[0] = this->PlayerPos.x;
			this->testCamera.p[1] = this->PlayerPos.y;
			this->testCamera.p[2] = this->PlayerPos.z;
			this->testCamera.lookAt[0] = temp.x;
			this->testCamera.lookAt[1] = temp.y;
			this->testCamera.lookAt[2] = temp.z;
		default:
			break;
	}
	eosPrepareCamera(this->testCamera);



	// init frame buffer (is there a cost to doing this every frame?)
	frameBuffer = (float *)malloc(this->testCamera.w*this->testCamera.h*3*sizeof(float));

	// call the renderer
	this->myInitData.rendererSettings->flags = this->rendererStates[this->currentRS]|(this->subsampleDebug?EOS_RENDERER_SUBSAMPLE_DEBUG:0);
	if (this->ambientOcclusionDebug) {
		this->myInitData.rendererSettings->flags |= EOS_RENDERER_AMBIENT_OCCLUSION;
	}
	// rendererSettings.flags &= ~EOS_RENDERER_MULTITHREADED; // uncomment to disable multithreading
	eosRenderScene(&this->testScene, &this->testCamera, frameBuffer, *(this->myInitData.rendererSettings));

	// track exposure
	eosSampleExposure(frameBuffer, testCamera.w, testCamera.h, &tracker);

	// apply filters - holy crap these are slow
	if (rendererStates[currentRS]&EOS_RENDERER_STANDARD) {
		// eosFilterExposure(2.0f, frameBuffer, testCamera.w, testCamera.h);
		// eosFilterGamma(2.2f, frameBuffer, testCamera.w, testCamera.h);
	}
	eosFilterSoften(2, 0.25, frameBuffer, testCamera.w, testCamera.h);
	eosFilterFastFloatExposure(tracker.currentExposure, frameBuffer, testCamera.w, testCamera.h);
	eosFilterFastFloatGamma(2.2f, frameBuffer, testCamera.w, testCamera.h);

	float * frameBufferResized = NULL;

	if ((this->testCamera.w==this->myInitData.screen->w) && (this->testCamera.h==this->myInitData.screen->h)) {
		// if the screen and internal buffer are the same size, do a simple copy
		frameBufferTemp = frameBuffer;
	} else {
		// if they're different sizes we need to do a stretch & copy
		frameBufferResized = (float *)malloc(this->myInitData.screen->w*this->myInitData.screen->h*3*sizeof(float));
		stbir_resize_float(frameBuffer, this->testCamera.w, this->testCamera.h, 0,
			frameBufferResized, this->myInitData.screen->w, this->myInitData.screen->h, 0, 3);
		frameBufferTemp = frameBufferResized;
	}

	vscrTemp = this->vscr;
	// iterate through each row
	for (j=0; j<this->myInitData.screen->h; j+=this->pixelStride) {
		// iterate thorugh each column
		for (i=0; i<this->myInitData.screen->w; i+=this->pixelStride) {
			vscrTemp[0] = (unsigned char)(255.0f*fmax(0.0f, fmin(1.0f, frameBufferTemp[0])));
			vscrTemp[1] = (unsigned char)(255.0f*fmax(0.0f, fmin(1.0f, frameBufferTemp[1])));
			vscrTemp[2] = (unsigned char)(255.0f*fmax(0.0f, fmin(1.0f, frameBufferTemp[2])));

			vscrTemp += 3;
			frameBufferTemp += 3;
		}
	}

	// free the frame buffer (again, cost to doing this every frame?)
	free(frameBuffer);
	if (frameBufferResized != NULL) free(frameBufferResized);

	// fast gamma filter
	if (!(rendererStates[currentRS]&EOS_RENDERER_STANDARD)) {
		// eosFilterFastExposure(2.0f, vscr, testCamera.w, testCamera.h);
		// eosFilterFastGamma(2.2, vscr, testCamera.w, testCamera.h);
	}

	SDL_BlitSurface(this->vscrSurface, NULL, this->myInitData.screen, NULL);


	// calculate framerate
	if (this->dt > 0) {
		sprintf_s(FrameTitle, 256, "EOS Tracer - %.2ffps", 1000.0f/(float)(this->dt));
	}
	SDL_WM_SetCaption(FrameTitle, NULL);

	// write text
	if (this->outputStats) {
		char 			outStr[1024];
		SDL_Color		fontColor={255, 255, 255};
		SDL_Surface		*text_surface;
		SDL_Rect		textDest;

		textDest.x = 8;
		textDest.y = 8;

		text_surface = TTF_RenderText_Blended(this->myFont, "Press left/right to change quality settings", fontColor);
		textDest.w = text_surface->w;
		textDest.h = text_surface->h;
		SDL_BlitSurface(text_surface, NULL, this->myInitData.screen, &textDest);
		SDL_FreeSurface(text_surface);
		textDest.y += TTF_FontLineSkip(this->myFont);
	
		text_surface = TTF_RenderText_Blended(this->myFont, "\"Z\": toggle ambient occlusion. \"X\": toggle subsample interpolation. \"C\": change camera.", fontColor);
		textDest.w = text_surface->w;
		textDest.h = text_surface->h;
		SDL_BlitSurface(text_surface, NULL, this->myInitData.screen, &textDest);
		SDL_FreeSurface(text_surface);
		textDest.y += TTF_FontLineSkip(this->myFont);
	
		if (rendererStates[currentRS]&EOS_RENDERER_STANDARD) {
			sprintf_s(outStr, 1024, "Renderer Mode: Standard (%.2ffps) %s", 1.0f/(float)dt,
				(rendererStates[currentRS]&EOS_RENDERER_ANTIALIAS)?" ANTIALIASED":"");
			text_surface = TTF_RenderText_Blended(this->myFont, outStr, fontColor);
			textDest.w = text_surface->w;
			textDest.h = text_surface->h;
			SDL_BlitSurface(text_surface, NULL, this->myInitData.screen, &textDest);
			SDL_FreeSurface(text_surface);
			textDest.y += TTF_FontLineSkip(this->myFont);
		} else {
			sprintf_s(outStr, 1024, "Renderer Mode: Subdivision (%.2ffps)", 1.0f/(float)dt);
			text_surface = TTF_RenderText_Blended(this->myFont, outStr, fontColor);
			textDest.w = text_surface->w;
			textDest.h = text_surface->h;
			SDL_BlitSurface(text_surface, NULL, this->myInitData.screen, &textDest);
			SDL_FreeSurface(text_surface);
			textDest.y += TTF_FontLineSkip(this->myFont);

			sprintf_s(outStr, 1024, "Subsample checks:%s%s%s", 
				(this->rendererStates[this->currentRS]&EOS_RENDERER_SUBSAMPLE_DIFFUSE)?" SHADOWS":"",
				(this->rendererStates[this->currentRS]&EOS_RENDERER_SUBSAMPLE_SPECULAR)?" SPECULAR":"",
				(this->rendererStates[this->currentRS]&EOS_RENDERER_SUBSAMPLE_REFLECTION)?" REFLECTIONS":"");
			text_surface = TTF_RenderText_Blended(this->myFont, outStr, fontColor);
			textDest.w = text_surface->w;
			textDest.h = text_surface->h;
			SDL_BlitSurface(text_surface, NULL, this->myInitData.screen, &textDest);
			SDL_FreeSurface(text_surface);
			textDest.y += TTF_FontLineSkip(this->myFont);

			sprintf_s(outStr, 1024, "Block Interpolation: %s", (subsampleDebug?"OFF":"ON"));
			text_surface = TTF_RenderText_Blended(this->myFont, outStr, fontColor);
			textDest.w = text_surface->w;
			textDest.h = text_surface->h;
			SDL_BlitSurface(text_surface, NULL, this->myInitData.screen, &textDest);
			SDL_FreeSurface(text_surface);
			textDest.y += TTF_FontLineSkip(this->myFont);
		}
	}

	SDL_Flip(this->myInitData.screen);
	SDL_Delay(1);
}




void				GSSampleScene::BuildTorus(eosRendererMesh &mesh, float r1, float r2, int steps1, int steps2) {
	int i, j;
	int currentTri = 0;
	float twoPi = 2.0f*3.1415926535897932384626433832795f; // copied from calc.exe
	float thetaInc, phiInc;

	thetaInc = twoPi/(float)(steps1);
	phiInc = twoPi/(float)(steps2);

	// Create 1 material
	mesh.nMats = 1;
	mesh.mats = (eosMaterial *)malloc(mesh.nMats*sizeof(eosMaterial));
	mesh.mats[0] = eosCreateMaterial(); // load the default material first
	mesh.mats[0].color[0] = 0.4f;
	mesh.mats[0].color[1] = 0.01f;
	mesh.mats[0].color[2] = 0.01f;
	if (checkerBoard.buff == NULL) {
		mesh.mats[0].tex = NULL;
	} else {
		mesh.mats[0].tex = &checkerBoard;
	}
	// mesh.mats[0].tex=NULL;
	mesh.mats[0].specular = 0.0f;
	mesh.mats[0].glossReflection = 0.5f;
	mesh.mats[0].fresnelTerm = 0.5f;
	mesh.mats[0].castsShadow = true;
	mesh.mats[0].unshaded=false;

	// Tris and other render info
	mesh.nTris = steps1*steps2*2;
	mesh.tris = (eosTriangle *)malloc(mesh.nTris*sizeof(eosTriangle));
	mesh.renderInfo = (eosTriangleRenderInfo *)malloc(mesh.nTris*sizeof(eosTriangleRenderInfo));
	for (j=0;j<steps1;j++) {
		for (i=0;i<steps2;i++) {
			// first triangle
			mesh.tris[currentTri].p1[0] = (r1 + r2*cos(phiInc*(i+0)))*cos(thetaInc*(j+0));
			mesh.tris[currentTri].p1[1] = r2*sin(phiInc*(i+0));
			mesh.tris[currentTri].p1[2] = (r1 + r2*cos(phiInc*(i+0)))*sin(thetaInc*(j+0));
			mesh.tris[currentTri].p2[0] = (r1 + r2*cos(phiInc*(i+1)))*cos(thetaInc*(j+0));
			mesh.tris[currentTri].p2[1] = r2*sin(phiInc*(i+1));
			mesh.tris[currentTri].p2[2] = (r1 + r2*cos(phiInc*(i+1)))*sin(thetaInc*(j+0));
			mesh.tris[currentTri].p3[0] = (r1 + r2*cos(phiInc*(i+0)))*cos(thetaInc*(j+1));
			mesh.tris[currentTri].p3[1] = r2*sin(phiInc*(i+0));
			mesh.tris[currentTri].p3[2] = (r1 + r2*cos(phiInc*(i+0)))*sin(thetaInc*(j+1));
			eosSetupTriangle(mesh.tris[currentTri]);

			mesh.renderInfo[currentTri].n1[0] = (cos(phiInc*(i+0)))*cos(thetaInc*(j+0));
			mesh.renderInfo[currentTri].n1[1] = sin(phiInc*(i+0));
			mesh.renderInfo[currentTri].n1[2] = (cos(phiInc*(i+0)))*sin(thetaInc*(j+0));
			mesh.renderInfo[currentTri].n2[0] = (cos(phiInc*(i+1)))*cos(thetaInc*(j+0));
			mesh.renderInfo[currentTri].n2[1] = sin(phiInc*(i+1));
			mesh.renderInfo[currentTri].n2[2] = (cos(phiInc*(i+1)))*sin(thetaInc*(j+0));
			mesh.renderInfo[currentTri].n3[0] = (cos(phiInc*(i+0)))*cos(thetaInc*(j+1));
			mesh.renderInfo[currentTri].n3[1] = sin(phiInc*(i+0));
			mesh.renderInfo[currentTri].n3[2] = (cos(phiInc*(i+0)))*sin(thetaInc*(j+1));

			mesh.renderInfo[currentTri].tc1[0] = 12.0f*(j+0)/(float)steps1;
			mesh.renderInfo[currentTri].tc1[1] = 6.0f*(i+0)/(float)steps2;
			mesh.renderInfo[currentTri].tc2[0] = 12.0f*(j+0)/(float)steps1;
			mesh.renderInfo[currentTri].tc2[1] = 6.0f*(i+1)/(float)steps2;
			mesh.renderInfo[currentTri].tc3[0] = 12.0f*(j+1)/(float)steps1;
			mesh.renderInfo[currentTri].tc3[1] = 6.0f*(i+0)/(float)steps2;

			mesh.renderInfo[currentTri].mat = &(mesh.mats[0]);
			mesh.renderInfo[currentTri].tracerGroup = (i<(steps2>>1)?2:0) | (j<(steps1>>1)?1:0);

			currentTri++;
			// second triangle
			mesh.tris[currentTri].p1[0] = (r1 + r2*cos(phiInc*(i+0)))*cos(thetaInc*(j+1));
			mesh.tris[currentTri].p1[1] = r2*sin(phiInc*(i+0));
			mesh.tris[currentTri].p1[2] = (r1 + r2*cos(phiInc*(i+0)))*sin(thetaInc*(j+1));
			mesh.tris[currentTri].p2[0] = (r1 + r2*cos(phiInc*(i+1)))*cos(thetaInc*(j+0));
			mesh.tris[currentTri].p2[1] = r2*sin(phiInc*(i+1));
			mesh.tris[currentTri].p2[2] = (r1 + r2*cos(phiInc*(i+1)))*sin(thetaInc*(j+0));
			mesh.tris[currentTri].p3[0] = (r1 + r2*cos(phiInc*(i+1)))*cos(thetaInc*(j+1));
			mesh.tris[currentTri].p3[1] = r2*sin(phiInc*(i+1));
			mesh.tris[currentTri].p3[2] = (r1 + r2*cos(phiInc*(i+1)))*sin(thetaInc*(j+1));
			eosSetupTriangle(mesh.tris[currentTri]);

			mesh.renderInfo[currentTri].n1[0] = (cos(phiInc*(i+0)))*cos(thetaInc*(j+1));
			mesh.renderInfo[currentTri].n1[1] = sin(phiInc*(i+0));
			mesh.renderInfo[currentTri].n1[2] = (cos(phiInc*(i+0)))*sin(thetaInc*(j+1));
			mesh.renderInfo[currentTri].n2[0] = (cos(phiInc*(i+1)))*cos(thetaInc*(j+0));
			mesh.renderInfo[currentTri].n2[1] = sin(phiInc*(i+1));
			mesh.renderInfo[currentTri].n2[2] = (cos(phiInc*(i+1)))*sin(thetaInc*(j+0));
			mesh.renderInfo[currentTri].n3[0] = (cos(phiInc*(i+1)))*cos(thetaInc*(j+1));
			mesh.renderInfo[currentTri].n3[1] = sin(phiInc*(i+1));
			mesh.renderInfo[currentTri].n3[2] = (cos(phiInc*(i+1)))*sin(thetaInc*(j+1));

			mesh.renderInfo[currentTri].tc1[0] = 12.0f*(j+1)/(float)steps1;
			mesh.renderInfo[currentTri].tc1[1] = 6.0f*(i+0)/(float)steps2;
			mesh.renderInfo[currentTri].tc2[0] = 12.0f*(j+0)/(float)steps1;
			mesh.renderInfo[currentTri].tc2[1] = 6.0f*(i+1)/(float)steps2;
			mesh.renderInfo[currentTri].tc3[0] = 12.0f*(j+1)/(float)steps1;
			mesh.renderInfo[currentTri].tc3[1] = 6.0f*(i+1)/(float)steps2;

			mesh.renderInfo[currentTri].mat = &(mesh.mats[0]);
			mesh.renderInfo[currentTri].tracerGroup = (i<(steps1>>2)?2:0) | (j<(steps1>>1)?1:0);

			currentTri++;
		}
	}

	// B-KD Tree
	mesh.tree = (eosTree *)malloc(sizeof(eosTree));
	eosBuildTree(mesh.tree, mesh.tris, mesh.nTris);
}

float *				GSSampleScene::LoadRawFloat(char *filename, int w, int h, bool flipRows) {
	FILE *fp = NULL;
	fopen_s(&fp, filename, "rb");
	float *buff;
	float *rowFlipper;
	if (fp) {
		buff = (float *)malloc(w*h*3*sizeof(float));
		fread((void *)buff, w*h*3*sizeof(float), 1, fp);
		fclose(fp);
		if (flipRows) {
			rowFlipper = (float *)malloc(w*3*sizeof(float));
			for (int i=0;i<h/2;i++) {
				memcpy(rowFlipper, (buff + i*w*3), w*3*sizeof(float));
				memcpy((buff + i*w*3), (buff + (h-i-1)*w*3), w*3*sizeof(float));
				memcpy((buff + (h-i-1)*w*3), rowFlipper, w*3*sizeof(float));
			}
			free(rowFlipper);
		}
		return buff;
	} else {
		return NULL;
	}
}

unsigned char *		GSSampleScene::LoadImage(char *filename, int &w, int &h) {
	SDL_Surface *input = IMG_Load(filename);
	unsigned char *buff;
	if (input) {
		SDL_SetAlpha(input, 0, SDL_ALPHA_OPAQUE);
		w = input->w;
		h = input->h;
		buff = (unsigned char *)malloc(w*h*4*sizeof(unsigned char));
		SDL_Surface *tmp = SDL_CreateRGBSurfaceFrom(buff, w, h, 32, w*4, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
		SDL_BlitSurface(input, NULL, tmp, NULL);
		SDL_FreeSurface(tmp);
		SDL_FreeSurface(input);
		return buff;
	} else {
		return NULL;
	}
}

eosTexture			GSSampleScene::LoadTextureFromSDL(char *filename) {
	eosTexture temp;
	int w, h;
	temp.buff = this->LoadImage(filename, w, h);
	if (temp.buff) {
		temp.w = w;
		temp.h = h;
		temp.planes = 4;
		temp.type = EOS_TEXTURE_BYTE;
		temp.filter = EOS_TEXTURE_NEAREST;
		temp.repeatX = temp.repeatY = EOS_TEXTURE_REPEAT;
	}
	return temp;
}

float				GSSampleScene::PerlinGetRand(int x, int y, int z) {
        int n = x + y*17 +z*31;
        n = (n << 13)^n;
        return ( 1.000f - ( (n * (n * n * 15731 + 789221) +
                1376312589) & 0x7FFFFFFF) / 1073741824.0f);
}

float				GSSampleScene::PerlinInterpolate(float a, float b, float pct) {

	const float pi = 3.1415926535897932384626433832795f;
	pct = 0.5f + 0.5f*cos(pi + pi*pct);
	return (b*pct + a*(1.0f-pct));
}

float				GSSampleScene::PerlinNoiseFunc(float x, float y, float z, float wl) {

		int px = (int)(x/wl);
        int py = (int)(y/wl);
        int pz = (int)(z/wl);

        x = (x-wl*px)/wl;
        y = (y-wl*py)/wl;
        z = (z-wl*pz)/wl;

		if (x<0) {x=1+x; px--;}
		if (y<0) {y=1+y; py--;}
		if (z<0) {z=1+z; pz--;}

		int px2 = px+1;
		int py2 = py+1;
		int pz2 = pz+1;

        float val[6];

        val[0] = this->PerlinInterpolate(this->PerlinGetRand(px, py, pz), this->PerlinGetRand(px2, py, pz), x);
        val[1] = this->PerlinInterpolate(this->PerlinGetRand(px, py2, pz), this->PerlinGetRand(px2, py2, pz), x);
        val[2] = this->PerlinInterpolate(this->PerlinGetRand(px, py, pz2), this->PerlinGetRand(px2, py, pz2), x);
        val[3] = this->PerlinInterpolate(this->PerlinGetRand(px, py2, pz2), this->PerlinGetRand(px2, py2, pz2), x);

        val[4] = this->PerlinInterpolate(val[0], val[1], y);
        val[5] = this->PerlinInterpolate(val[2], val[3], y);

        return this->PerlinInterpolate(val[4], val[5], z);
}

float				GSSampleScene::PerlinGetXYZ(float x, float y, float z, float persistence) {

	float accumulator = 0.0f;
	float persist = 1.0f;
	float wl = 1.0f;

	for (int i=0; i<8; i++) {
		accumulator += persist*this->PerlinNoiseFunc(x, y, z, wl);
		wl /= 2.0;
		persist *= persistence;
	}

	accumulator = 0.5f + 0.25f*accumulator;

	return accumulator;
}

void				GSSampleScene::GenerateHeightmap(void) {

	int heightmapW = 257;
	int heightmapH = 257;
	int heightmapElev = 100;
	int i, j;
	unsigned int offs;
	const int shifter = 4;

	SVect3d *heightmapP = (SVect3d *)malloc(heightmapW*heightmapH*sizeof(SVect3d));
	SVect3d *heightmapN = (SVect3d *)malloc(heightmapW*heightmapH*sizeof(SVect3d));
	
	offs = 0;
	for (j=0;j<heightmapH;j++) {
		for (i=0;i<heightmapW;i++) {
			heightmapP[offs] = v3d((float)(i-(heightmapW>>1)), -(this->PerlinGetXYZ(i*0.01f, j*0.01f, 100.0f, 0.5f)-0.5f)*heightmapElev ,-(float)(j-(heightmapW>>1)));
			heightmapN[offs] = v3d(0.0f, 1.0f, 0.0f);
			offs ++;
		}
	}
	/*
	for (j=0;j<heightmapH-1;j++) {
		offs = heightmapW*j;
		for (i=0;i<heightmapW-1;i++) {
			heightmapP[offs].y = (heightmapP[offs].y + heightmapP[offs+1].y + heightmapP[offs+heightmapW].y + heightmapP[offs+heightmapW+1].y)*0.25f;
			offs++;
		}
	}
	*/
	// normals across top & bottom row
	for (i=1;i<heightmapW-1;i++) {
		offs = i;
		heightmapN[offs] = vmCrossProduct(vmNormalize(heightmapP[offs+1]-heightmapP[offs-1]), v3d(0, 0, -1));
		offs = (heightmapH-1)*heightmapW + i;
		heightmapN[offs] = vmCrossProduct(vmNormalize(heightmapP[offs+1]-heightmapP[offs-1]), v3d(0, 0, -1));
	}

	// normals across left and right
	for (j=1;j<heightmapH-1;j++) {
		offs = (j*heightmapW);
		heightmapN[offs] = vmCrossProduct(v3d(1, 0, 0), vmNormalize(heightmapP[offs+heightmapW]-heightmapP[offs-heightmapW]));
		offs = (j*heightmapW) + (heightmapW-1);
		heightmapN[offs] = vmCrossProduct(v3d(1, 0, 0), vmNormalize(heightmapP[offs+heightmapW]-heightmapP[offs-heightmapW]));
	}

	// all the other normals
	for (j=1;j<heightmapH-1;j++) {
		for (i=1;i<heightmapW-1;i++) {
			offs = j*heightmapW + i;
			heightmapN[offs] = vmCrossProduct(vmNormalize(heightmapP[offs+1]-heightmapP[offs-1]), vmNormalize(heightmapP[offs+heightmapW]-heightmapP[offs-heightmapW]));
			heightmapN[offs] = vmNormalize(heightmapN[offs]);
		}
	}

	// now build the actual stuff
	this->heightMap.nTris = (heightmapW-1)*(heightmapH-1)*2;
	this->heightMap.tris = (eosTriangle *)malloc(this->heightMap.nTris*sizeof(eosTriangle));
	this->heightMap.renderInfo = (eosTriangleRenderInfo *)malloc(this->heightMap.nTris*sizeof(eosTriangleRenderInfo));
	this->heightMap.mats = (eosMaterial *)malloc(sizeof(eosMaterial));
	
	this->heightMap.mats[0] = eosCreateMaterial();
	this->heightMap.mats[0].color[0] = 0.25f;
	this->heightMap.mats[0].color[1] = 0.10f;
	this->heightMap.mats[0].color[2] = 0.05f;
	this->heightMap.mats[0].tex = NULL;
	
	offs = 0;
	for (j=0;j<(heightmapH-1);j++) {
		for (i=0;i<(heightmapW-1);i++) {
			// first triangle
			this->heightMap.tris[offs].p1[0] = heightmapP[(j+0)*heightmapW+(i+0)].x;
			this->heightMap.tris[offs].p1[1] = heightmapP[(j+0)*heightmapW+(i+0)].y;
			this->heightMap.tris[offs].p1[2] = heightmapP[(j+0)*heightmapW+(i+0)].z;
			this->heightMap.tris[offs].p2[0] = heightmapP[(j+0)*heightmapW+(i+1)].x;
			this->heightMap.tris[offs].p2[1] = heightmapP[(j+0)*heightmapW+(i+1)].y;
			this->heightMap.tris[offs].p2[2] = heightmapP[(j+0)*heightmapW+(i+1)].z;
			this->heightMap.tris[offs].p3[0] = heightmapP[(j+1)*heightmapW+(i+0)].x;
			this->heightMap.tris[offs].p3[1] = heightmapP[(j+1)*heightmapW+(i+0)].y;
			this->heightMap.tris[offs].p3[2] = heightmapP[(j+1)*heightmapW+(i+0)].z;

			eosSetupTriangle(this->heightMap.tris[offs]);
			
			this->heightMap.renderInfo[offs].n1[0] = heightmapN[(j+0)*heightmapW+(i+0)].x;
			this->heightMap.renderInfo[offs].n1[1] = heightmapN[(j+0)*heightmapW+(i+0)].y;
			this->heightMap.renderInfo[offs].n1[2] = heightmapN[(j+0)*heightmapW+(i+0)].z;
			this->heightMap.renderInfo[offs].n2[0] = heightmapN[(j+0)*heightmapW+(i+1)].x;
			this->heightMap.renderInfo[offs].n2[1] = heightmapN[(j+0)*heightmapW+(i+1)].y;
			this->heightMap.renderInfo[offs].n2[2] = heightmapN[(j+0)*heightmapW+(i+1)].z;
			this->heightMap.renderInfo[offs].n3[0] = heightmapN[(j+1)*heightmapW+(i+0)].x;
			this->heightMap.renderInfo[offs].n3[1] = heightmapN[(j+1)*heightmapW+(i+0)].y;
			this->heightMap.renderInfo[offs].n3[2] = heightmapN[(j+1)*heightmapW+(i+0)].z;

			this->heightMap.renderInfo[offs].tc1[0] = (float)(i+0);
			this->heightMap.renderInfo[offs].tc1[1] = (float)(j+0);
			this->heightMap.renderInfo[offs].tc2[0] = (float)(i+1);
			this->heightMap.renderInfo[offs].tc2[1] = (float)(j+0);
			this->heightMap.renderInfo[offs].tc3[0] = (float)(i+0);
			this->heightMap.renderInfo[offs].tc3[1] = (float)(j+1);

			this->heightMap.renderInfo[offs].mat = &(this->heightMap.mats[0]);
			this->heightMap.renderInfo[offs].tracerGroup = (j & (0xFFFF<<shifter)) | (i>>shifter);

			offs++;

			// second triangle
			this->heightMap.tris[offs].p1[0] = heightmapP[(j+1)*heightmapW+(i+0)].x;
			this->heightMap.tris[offs].p1[1] = heightmapP[(j+1)*heightmapW+(i+0)].y;
			this->heightMap.tris[offs].p1[2] = heightmapP[(j+1)*heightmapW+(i+0)].z;
			this->heightMap.tris[offs].p2[0] = heightmapP[(j+0)*heightmapW+(i+1)].x;
			this->heightMap.tris[offs].p2[1] = heightmapP[(j+0)*heightmapW+(i+1)].y;
			this->heightMap.tris[offs].p2[2] = heightmapP[(j+0)*heightmapW+(i+1)].z;
			this->heightMap.tris[offs].p3[0] = heightmapP[(j+1)*heightmapW+(i+1)].x;
			this->heightMap.tris[offs].p3[1] = heightmapP[(j+1)*heightmapW+(i+1)].y;
			this->heightMap.tris[offs].p3[2] = heightmapP[(j+1)*heightmapW+(i+1)].z;
			
			eosSetupTriangle(this->heightMap.tris[offs]);

			this->heightMap.renderInfo[offs].n1[0] = heightmapN[(j+1)*heightmapW+(i+0)].x;
			this->heightMap.renderInfo[offs].n1[1] = heightmapN[(j+1)*heightmapW+(i+0)].y;
			this->heightMap.renderInfo[offs].n1[2] = heightmapN[(j+1)*heightmapW+(i+0)].z;
			this->heightMap.renderInfo[offs].n2[0] = heightmapN[(j+0)*heightmapW+(i+1)].x;
			this->heightMap.renderInfo[offs].n2[1] = heightmapN[(j+0)*heightmapW+(i+1)].y;
			this->heightMap.renderInfo[offs].n2[2] = heightmapN[(j+0)*heightmapW+(i+1)].z;
			this->heightMap.renderInfo[offs].n3[0] = heightmapN[(j+1)*heightmapW+(i+1)].x;
			this->heightMap.renderInfo[offs].n3[1] = heightmapN[(j+1)*heightmapW+(i+1)].y;
			this->heightMap.renderInfo[offs].n3[2] = heightmapN[(j+1)*heightmapW+(i+1)].z;

			this->heightMap.renderInfo[offs].tc1[0] = (float)(i+0);
			this->heightMap.renderInfo[offs].tc1[1] = (float)(j+1);
			this->heightMap.renderInfo[offs].tc2[0] = (float)(i+1);
			this->heightMap.renderInfo[offs].tc2[1] = (float)(j+0);
			this->heightMap.renderInfo[offs].tc3[0] = (float)(i+1);
			this->heightMap.renderInfo[offs].tc3[1] = (float)(j+1);

			this->heightMap.renderInfo[offs].mat = &(this->heightMap.mats[0]);
			this->heightMap.renderInfo[offs].tracerGroup = (j & (0xFFFF<<shifter)) | (i>>shifter);

			offs++;
		}
	}

	free(heightmapP);
	free(heightmapN);

	// build tree
	this->heightMap.tree = (eosTree *)malloc(sizeof(eosTree));
	eosBuildTree(this->heightMap.tree, this->heightMap.tris, this->heightMap.nTris);
}
