#ifndef _SAMPLESCENE_H
#define _SAMPLESCENE_H

#include "SDL.h"						// SDL library
#include "SDL_image.h"					// SDL Image library (for image file IO)
#include "SDL_ttf.h"					// SDL TrueType library (for putting text on the screen)

#include "../eos/eos.h"
#include "../gameState.h"

struct SampleSceneInitData {
	int						DefaultScreenWidth;
	int						DefaultScreenHeight;
	int						InternalWidth;
	int						InternalHeight;
	SDL_Surface *			screen;
	float					PlayerMouseSensitivity;
	float					PlayerKeyboardSensitivity;
	int *					gameRunning;
	eosRendererSettings *	rendererSettings;
	bool					FullScreen;
};

struct SampleSceneCycleData {
};

class GSSampleScene:public GameState {
	private:
		const int				nCameras;
		const int				nRendererStates;

		unsigned int *			rendererStates;

		SDL_Surface *			vscrSurface;
		unsigned char *			vscr;

		float					AnimTimer;
		int						pixelStride;
		int						currentRS;
		eosCamera				testCamera;
		eosScene				testScene;
		eosLight *				tmpLight;
		eosLight *				ambientLight;
		eosLight *				localLight;
		eosRendererMesh			torusMesh, torusMesh2, heightMap, loadedMesh;
		eosEnvironmentMap		env;
		eosTexture				checkerBoard, sand;
		bool					subsampleDebug;
		bool					ambientOcclusionDebug;
		int						currentCamera;
		eosExposureTracker		tracker;
		TTF_Font *				myFont;
		float					dt;

		SVect3d					PlayerPos;
		float					PlayerHeading;
		float					PlayerPitch;
		bool					outputStats;

		SampleSceneInitData		myInitData;
		int						FullScreenWidth;
		int						FullScreenHeight;

		void				GSSampleScene::BuildTorus(eosRendererMesh &mesh, float r1, float r2, int steps1, int steps2);
		float *				LoadRawFloat(char *filename, int w, int h, bool flipRows = false);
		unsigned char *		LoadImage(char *filename, int &w, int &h);
		eosTexture			LoadTextureFromSDL(char *filename);
		float				PerlinGetRand(int x, int y, int z);
		float				PerlinInterpolate(float a, float b, float pct);
		float				PerlinNoiseFunc(float x, float y, float z, float wl);
		float				PerlinGetXYZ(float x, float y, float z, float persistence);
		void				GenerateHeightmap(void);

		inline float		fmin(const float &a, const float &b) { return (a<b?a:b);};
		inline float		fmax(const float &a, const float &b) { return (a>b?a:b);};


	public:
		void				SetFullScreen(bool inFullScreen);
		inline bool			GetFullScreen(void) {return this->myInitData.FullScreen;};

		void Init(void *initData);
		void Load(void);
		void Unload(void);
		void Cycle(float dt, void *cycleData);
		void Draw();

		GSSampleScene():
			nCameras(4), nRendererStates(6), AnimTimer(0), pixelStride(1), currentRS(4), subsampleDebug(false), ambientOcclusionDebug(false), currentCamera(0), myFont(NULL), dt(0.00001f),FullScreenWidth(0),FullScreenHeight(0),vscr(NULL),vscrSurface(NULL),outputStats(true)
		{
			this->rendererStates = new unsigned int [this->nRendererStates];
			this->rendererStates[0] = EOS_RENDERER_MULTITHREADED|EOS_RENDERER_SUBSAMPLE;
			this->rendererStates[1] = EOS_RENDERER_MULTITHREADED|EOS_RENDERER_SUBSAMPLE|EOS_RENDERER_SUBSAMPLE_DIFFUSE;
			this->rendererStates[2] = EOS_RENDERER_MULTITHREADED|EOS_RENDERER_SUBSAMPLE|EOS_RENDERER_SUBSAMPLE_DIFFUSE|EOS_RENDERER_SUBSAMPLE_SPECULAR;
			this->rendererStates[3] = EOS_RENDERER_MULTITHREADED|EOS_RENDERER_SUBSAMPLE|EOS_RENDERER_SUBSAMPLE_DIFFUSE|EOS_RENDERER_SUBSAMPLE_SPECULAR|EOS_RENDERER_SUBSAMPLE_REFLECTION;
			this->rendererStates[4] = EOS_RENDERER_MULTITHREADED|EOS_RENDERER_STANDARD|EOS_RENDERER_JITTER_SOFT_SHADOWS;
			this->rendererStates[5] = EOS_RENDERER_MULTITHREADED|EOS_RENDERER_STANDARD|EOS_RENDERER_JITTER_SOFT_SHADOWS|EOS_RENDERER_ANTIALIAS;
		};

		~GSSampleScene() {
			if (this->vscr) {
				SDL_FreeSurface(this->vscrSurface);
				free(this->vscr);
			}
			delete [] this->rendererStates;
			this->Unload();	// user can call this explicitly, but we'll call it here just in case they didn't
		};
};

#endif // _SAMPLESCENE_H